/*
 * Copyright 2013 Mic
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include "emu6502.h"
#include "sidmapper.h"

#define UPDATE_NZ(val) regF &= ~(EMU6502_FLAG_Z | EMU6502_FLAG_N); \
				       regF |= ((uint8_t)val == 0) ? EMU6502_FLAG_Z : 0; \
				       regF |= (val & 0x80)

#define UPDATE_NZC(val, val2, sub) regF &= ~(EMU6502_FLAG_Z | EMU6502_FLAG_N | EMU6502_FLAG_C); \
				       regF |= ((uint8_t)val == (uint8_t)val2) ? EMU6502_FLAG_Z : 0; \
				       regF |= ((val - sub) & 0x80); \
				       regF |= (val >= val2) ? EMU6502_FLAG_C : 0

// == Arithmetic ==

#define ADC(val) temp16 = (uint16_t)(val) + (uint16_t)regA + ((regF & EMU6502_FLAG_C) ? 1 : 0); \
				 UPDATE_NZC(temp16, 0x100, 0); \
				 regF &= ~EMU6502_FLAG_V; \
				 regF |= ((!((regA ^ val) & 0x80)) && ((regA ^ temp16) & 0x80)) ? EMU6502_FLAG_V : 0; \
				 regA = (uint8_t)temp16

#define ASL(val) regF &= ~EMU6502_FLAG_C; \
	             regF |= (val & 0x80) ? EMU6502_FLAG_C : 0; \
	             val <<= 1; \
	             UPDATE_NZ(val)

#define BIT(val) regF &= ~(EMU6502_FLAG_Z | EMU6502_FLAG_V | EMU6502_FLAG_N); \
				 regF |= (((uint8_t)val & regA) == 0) ? EMU6502_FLAG_Z : 0; \
				 regF |= (val & 0xC0)

#define LSR(val) regF &= ~EMU6502_FLAG_C; \
	             regF |= (val & 0x01) ? EMU6502_FLAG_C : 0; \
	             val >>= 1; \
	             UPDATE_NZ(val)

#define ROL(val) temp8 = (val << 1) | ((regF & EMU6502_FLAG_C) ? 1 : 0); \
				 regF &= ~EMU6502_FLAG_C; \
	             regF |= (val & 0x80) ? EMU6502_FLAG_C : 0; \
	             val = temp8; \
	             UPDATE_NZ(val)

#define ROR(val) temp8 = (val >> 1) | ((regF & EMU6502_FLAG_C) ? 0x80 : 0); \
				 regF &= ~EMU6502_FLAG_C; \
	             regF |= (val & 0x01) ? EMU6502_FLAG_C : 0; \
	             val = temp8; \
	             UPDATE_NZ(val)

// == Addressing ==

// abs
// Doesn't update PC
#define ABS_ADDR() (sidMapper_readByte(regPC) + ((uint16_t)sidMapper_readByte(regPC+1) << 8))

// abs,X
// Updates PC
#define ABSX_ADDR(dest) dest = (sidMapper_readByte(regPC) + ((uint16_t)sidMapper_readByte(regPC+1) << 8)); \
	                    if ((dest & 0x100) != ((dest + regX) & 0x100)) cpuCycles++; \
	                    regPC += 2; \
						dest += regX

// abs,Y
// Updates PC
#define ABSY_ADDR(dest) dest = (sidMapper_readByte(regPC) + ((uint16_t)sidMapper_readByte(regPC+1) << 8)); \
	                    if ((dest & 0x100) != ((dest + regY) & 0x100)) cpuCycles++; \
	                    regPC += 2; \
						dest += regY

// zp
// Updates PC
#define ZP_ADDR() (sidMapper_readByte(regPC++))

// zp,X
// Updates PC
#define ZPX_ADDR() ((ZP_ADDR() + regX) & 0xFF)

// zp,Y
// Updates PC
#define ZPY_ADDR() ((ZP_ADDR() + regY) & 0xFF)

// (zp,X)
// Updates PC
#define INDX_ADDR(dest) dest = ZPX_ADDR(); \
						dest = (sidMapper_readByte(dest) + ((uint16_t)sidMapper_readByte((dest & 0xFF00)+((dest+1)&0xFF)) << 8))

// (zp),Y
// Updates PC
#define INDY_ADDR(dest) dest = ZP_ADDR(); \
						dest = (sidMapper_readByte(dest) + ((uint16_t)sidMapper_readByte((dest & 0xFF00)+((dest+1)&0xFF)) << 8)); \
						if ((dest & 0x100) != ((dest + regY) & 0x100)) cpuCycles++; \
						dest += regY


// ====

#define COND_BRANCH(cc) if ((cc)) { relAddr = sidMapper_readByte(regPC++); \
								    addr = regPC + relAddr; \
                                    cpuCycles += ((addr & 0x100) == (regPC & 0x100)) ? 1 : 2; \
                                    regPC = addr; } else { regPC++; } \
                                  cpuCycles += 2

#define PUSHB(val) sidMapper_writeByte(0x100 + (uint16_t)regS, val); \
	               regS--

#define PUSHW(val) sidMapper_writeByte(0x100 + (uint16_t)regS, ((uint16_t)val >> 8)); \
	               regS--; \
	               sidMapper_writeByte(0x100 + (uint16_t)regS, (uint8_t)val); \
	               regS--

#define PULLB(dest) regS++; \
	                dest = sidMapper_readByte(0x100 + (uint16_t)regS)


#define ILLEGAL_OP() cpuCycles += 2; \
					 regPC++

// ====

uint8_t regA, regX, regY, regS, regF;
uint16_t regPC;
uint32_t cpuCycles;
static uint16_t brkVector;


void emu6502_irq(uint16_t vector)
 {
	PUSHW(regPC);
	PUSHB(regF);
	regF |= EMU6502_FLAG_I;
	regPC = sidMapper_readByte(vector);
	regPC |= (uint16_t)sidMapper_readByte(vector + 1) << 8;
}

void emu6502_setBrkVector(uint16_t vector)
{
    brkVector = vector;
}


void emu6502_reset()
{
	brkVector = 0xfffe;
	regF = 0;
}


void emu6502_run(uint32_t maxCycles)
{
	uint16_t addr, temp16;
	uint8_t operand, temp8;
    int8_t relAddr;

	while (cpuCycles < maxCycles) {

		uint8_t opcode = sidMapper_readByte(regPC++);

		switch (opcode) {

// == ADC ==
		case 0x69:		// ADC imm
			operand = sidMapper_readByte(regPC++);
			ADC(operand);
			cpuCycles += 2;
			break;

		case 0x65:		// ADC zp
			operand = sidMapper_readByte(ZP_ADDR());
			ADC(operand);
			cpuCycles += 3;
			break;

		case 0x75:		// ADC zp,X
			operand = sidMapper_readByte(ZPX_ADDR());
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0x6D:		// ADC abs
			operand = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0x7D:		// ADC abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0x79:		// ADC abs,Y
			ABSY_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0x61:		// ADC (zp,X)
			INDX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ADC(operand);
			cpuCycles += 6;
			break;

		case 0x71:		// ADC (zp),Y
			INDY_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ADC(operand);
			cpuCycles += 5;
			break;


// == AND ==
		case 0x29:		// AND imm
			regA &= sidMapper_readByte(regPC++);
			UPDATE_NZ(regA);
			cpuCycles += 2;
			break;

		case 0x25:		// AND zp
			regA &= sidMapper_readByte(ZP_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 3;
			break;

		case 0x35:		// AND zp,X
			regA &= sidMapper_readByte(ZPX_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x2D:		// AND abs
			regA &= sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x3D:		// AND abs,X
			ABSX_ADDR(addr);
			regA &= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x39:		// AND abs,Y
			ABSY_ADDR(addr);
			regA &= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x21:		// AND (zp,X)
			INDX_ADDR(addr);
			regA &= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 6;
			break;

		case 0x31:		// AND (zp),Y
			INDY_ADDR(addr);
			regA &= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 5;
			break;


// == ASL ==
		case 0x0A:		// ASL A
			ASL(regA);
			cpuCycles += 2;
			break;

		case 0x06:		// ASL zp
			addr = ZP_ADDR();
			operand = sidMapper_readByte(addr);
			ASL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 5;
			break;

		case 0x16:		// ASL zp,X
			addr = ZPX_ADDR();
			operand = sidMapper_readByte(addr);
			ASL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x0E:		// ASL abs
			addr = ABS_ADDR();
			regPC += 2;
			operand = sidMapper_readByte(addr);
			ASL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x1E:		// ASL abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ASL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 7;
			break;


// == Bxx ==
		case 0x10:		// BPL rel
			COND_BRANCH((regF & EMU6502_FLAG_N) == 0);
			break;
		case 0x30:		// BMI rel
			COND_BRANCH(regF & EMU6502_FLAG_N);
			break;

		case 0x50:		// BVC rel
			COND_BRANCH((regF & EMU6502_FLAG_V) == 0);
			break;
		case 0x70:		// BVS rel
			COND_BRANCH(regF & EMU6502_FLAG_V);
			break;

		case 0x90:		// BCC rel
			COND_BRANCH((regF & EMU6502_FLAG_C) == 0);
			break;
		case 0xB0:		// BCS rel
			COND_BRANCH(regF & EMU6502_FLAG_C);
			break;

		case 0xD0:		// BNE rel
			COND_BRANCH((regF & EMU6502_FLAG_Z) == 0);
			break;
		case 0xF0:		// BEQ rel
			COND_BRANCH(regF & EMU6502_FLAG_Z);
			break;

// == BIT ==
		case 0x24:		// BIT zp
			operand = sidMapper_readByte(ZP_ADDR());
			BIT(operand);
			cpuCycles += 3;
			break;

		case 0x2C:		// BIT abs
			operand = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			BIT(operand);
			cpuCycles += 4;
			break;


// ====
		case 0x00:		// BRK
			regPC++;
			PUSHW(regPC);
			temp8 = regF | (EMU6502_FLAG_B | 0x20);
			PUSHB(temp8);
			regF |= (EMU6502_FLAG_B | EMU6502_FLAG_I);
			regPC = sidMapper_readByte(brkVector);
			regPC |= (uint16_t)sidMapper_readByte(brkVector+1) << 8;
			cpuCycles += 7;
			break;


// == CLx ==
		case 0x18:		// CLC
			regF &= ~EMU6502_FLAG_C;
			cpuCycles += 2;
			break;

		case 0xD8:		// CLD
			regF &= ~EMU6502_FLAG_D;
			cpuCycles += 2;
			break;

		case 0x58:		// CLI
			regF &= ~EMU6502_FLAG_I;
			cpuCycles += 2;
			break;

		case 0xB8:		// CLV
			regF &= ~EMU6502_FLAG_V;
			cpuCycles += 2;
			break;


// == CMP ==
		case 0xC9:		// CMP imm
			operand = sidMapper_readByte(regPC++);
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 2;
			break;

		case 0xC5:		// CMP zp
			operand = sidMapper_readByte(ZP_ADDR());
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 3;
			break;

		case 0xD5:		// CMP zp,X
			operand = sidMapper_readByte(ZPX_ADDR());
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 4;
			break;

		case 0xCD:		// CMP abs
			operand = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 4;
			break;

		case 0xDD:		// CMP abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 4;
			break;

		case 0xD9:		// CMP abs,Y
			ABSY_ADDR(addr);
			operand = sidMapper_readByte(addr);
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 4;
			break;

		case 0xC1:		// CMP (zp,X)
			INDX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 6;
			break;

		case 0xD1:		// CMP (zp),Y
			INDY_ADDR(addr);
			operand = sidMapper_readByte(addr);
			UPDATE_NZC(regA, operand, operand);
			cpuCycles += 5;
			break;

// == CPX ==
		case 0xE0:		// CPX imm
			operand = sidMapper_readByte(regPC++);
			UPDATE_NZC(regX, operand, operand);
			cpuCycles += 2;
			break;

		case 0xE4:		// CPX zp
			operand = sidMapper_readByte(ZP_ADDR());
			UPDATE_NZC(regX, operand, operand);
			cpuCycles += 3;
			break;

		case 0xEC:		// CPX abs
			operand = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZC(regX, operand, operand);
			cpuCycles += 4;
			break;

// == CPY ==
		case 0xC0:		// CPY imm
			operand = sidMapper_readByte(regPC++);
			UPDATE_NZC(regY, operand, operand);
			cpuCycles += 2;
			break;

		case 0xC4:		// CPY zp
			operand = sidMapper_readByte(ZP_ADDR());
			UPDATE_NZC(regY, operand, operand);
			cpuCycles += 3;
			break;

		case 0xCC:		// CPY abs
			operand = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZC(regY, operand, operand);
			cpuCycles += 4;
			break;


// == DEC ==
		case 0xC6:		// DEC zp
			addr = ZP_ADDR();
			operand = sidMapper_readByte(addr) - 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 5;
			break;

		case 0xD6:		// DEC zp,X
			addr = ZPX_ADDR();
			operand = sidMapper_readByte(addr) - 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0xCE:		// DEC abs
			addr = ABS_ADDR();
			regPC += 2;
			operand = sidMapper_readByte(addr) - 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0xDE:		// DEC abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr) - 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

// ====
		case 0xCA:		// DEX
			regX--;
			UPDATE_NZ(regX);
			cpuCycles += 2;
			break;

		case 0x88:		// DEY
			regY--;
			UPDATE_NZ(regY);
			cpuCycles += 2;
			break;


// == EOR ==
		case 0x49:		// EOR imm
			regA ^= sidMapper_readByte(regPC++);
			UPDATE_NZ(regA);
			cpuCycles += 2;
			break;

		case 0x45:		// EOR zp
			regA ^= sidMapper_readByte(ZP_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 3;
			break;

		case 0x55:		// EOR zp,X
			regA ^= sidMapper_readByte(ZPX_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x4D:		// EOR abs
			regA ^= sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x5D:		// EOR abs,X
			ABSX_ADDR(addr);
			regA ^= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x59:		// EOR abs,Y
			ABSY_ADDR(addr);
			regA ^= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x41:		// EOR (zp,X)
			INDX_ADDR(addr);
			regA ^= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 6;
			break;

		case 0x51:		// EOR (zp),Y
			INDY_ADDR(addr);
			regA ^= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 6;
			break;

// == INC ==
		case 0xE6:		// INC zp
			addr = ZP_ADDR();
			operand = sidMapper_readByte(addr) + 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 5;
			break;

		case 0xF6:		// INC zp,X
			addr = ZPX_ADDR();
			operand = sidMapper_readByte(addr) + 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0xEE:		// INC abs
			addr = ABS_ADDR();
			regPC += 2;
			operand = sidMapper_readByte(addr) + 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0xFE:		// INC abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr) + 1;
			UPDATE_NZ(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

// ====
		case 0xE8:		// INX
			regX++;
			UPDATE_NZ(regX);
			cpuCycles += 2;
			break;

		case 0xC8:		// INY
			regY++;
			UPDATE_NZ(regY);
			cpuCycles += 2;
			break;

// ====
		case 0x4C:		// JMP abs
			temp16 = regPC - 1;
			addr = sidMapper_readByte(regPC++);
			addr |= (uint16_t)sidMapper_readByte(regPC) << 8;
			regPC = addr;
			if (addr == temp16) {
				// Exit early for infinite loops
				return;
			}
			cpuCycles += 3;
			break;

		case 0x6C:		// JMP (abs)
			addr = ABS_ADDR();
			regPC = sidMapper_readByte(addr);
			regPC |= (uint16_t)sidMapper_readByte((addr&0xFF00)+((addr+1)&0xFF)) << 8;
			cpuCycles += 5;
			break;

		case 0x20:		// JSR abs
			 addr = sidMapper_readByte(regPC++);
			 addr |= (uint16_t)sidMapper_readByte(regPC) << 8;
			 PUSHW(regPC);
			 regPC = addr;
			 cpuCycles += 6;
			 break;

 // == LAX ==
		case 0xA7:		// LAX zp
			regA = sidMapper_readByte(ZP_ADDR());
			regX = regA;
			UPDATE_NZ(regA);
			cpuCycles += 3;
			break;

		case 0xB7:		// LAX zp,Y
			regA = sidMapper_readByte(ZPY_ADDR());
			regX = regA;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xAF:		// LAX abs
			regA = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			regX = regA;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xBF:		// LAX abs,Y
			ABSY_ADDR(addr);
			regA = sidMapper_readByte(addr);
			regX = regA;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xB3:		// LAX (zp),Y
			INDY_ADDR(addr);
			regA = sidMapper_readByte(addr);
			regX = regA;
			UPDATE_NZ(regA);
			cpuCycles += 5;
			break;

// == LDA ==
		case 0xA9:		// LDA imm
			regA = sidMapper_readByte(regPC++);
			UPDATE_NZ(regA);
			cpuCycles += 2;
			break;

		case 0xA5:		// LDA zp
			regA = sidMapper_readByte(ZP_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 3;
			break;

		case 0xB5:		// LDA zp,X
			regA = sidMapper_readByte(ZPX_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xAD:		// LDA abs
			regA = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xBD:		// LDA abs,X
			ABSX_ADDR(addr);
			regA = sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xB9:		// LDA abs,Y
			ABSY_ADDR(addr);
			regA = sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0xA1:		// LDA (zp,X)
			INDX_ADDR(addr);
			regA = sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 6;
			break;

		case 0xB1:		// LDA (zp),Y
			INDY_ADDR(addr);
			regA = sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 5;
			break;

// == LDX ==
		case 0xA2:		// LDX imm
			regX = sidMapper_readByte(regPC++);
			UPDATE_NZ(regX);
			cpuCycles += 2;
			break;

		case 0xA6:		// LDX zp
			regX = sidMapper_readByte(ZP_ADDR());
			UPDATE_NZ(regX);
			cpuCycles += 3;
			break;

		case 0xB6:		// LDX zp,Y
			regX = sidMapper_readByte(ZPY_ADDR());
			UPDATE_NZ(regX);
			cpuCycles += 4;
			break;

		case 0xAE:		// LDX abs
			regX = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZ(regX);
			cpuCycles += 4;
			break;

		case 0xBE:		// LDX abs,Y
			ABSY_ADDR(addr);
			regX = sidMapper_readByte(addr);
			UPDATE_NZ(regX);
			cpuCycles += 4;
			break;


// == LDY ==
		case 0xA0:		// LDY imm
			regY = sidMapper_readByte(regPC++);
			UPDATE_NZ(regY);
			cpuCycles += 2;
			break;

		case 0xA4:		// LDY zp
			regY = sidMapper_readByte(ZP_ADDR());
			UPDATE_NZ(regY);
			cpuCycles += 3;
			break;

		case 0xB4:		// LDY zp,X
			regY = sidMapper_readByte(ZPX_ADDR());
			UPDATE_NZ(regY);
			cpuCycles += 4;
			break;

		case 0xAC:		// LDY abs
			regY = sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZ(regY);
			cpuCycles += 4;
			break;

		case 0xBC:		// LDY abs,X
			ABSX_ADDR(addr);
			regY = sidMapper_readByte(addr);
			UPDATE_NZ(regY);
			cpuCycles += 4;
			break;


// == LSR ==
		case 0x4A:		// LSR A
			LSR(regA);
			cpuCycles += 2;
			break;

		case 0x46:		// LSR zp
			addr = ZP_ADDR();
			operand = sidMapper_readByte(addr);
			LSR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 5;
			break;

		case 0x56:		// LSR zp,X
			addr = ZPX_ADDR();
			operand = sidMapper_readByte(addr);
			LSR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x4E:		// LSR abs
			addr = ABS_ADDR();
			regPC += 2;
			operand = sidMapper_readByte(addr);
			LSR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x5E:		// LSR abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			LSR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 7;
			break;


// == ORA ==
		case 0x09:		// ORA imm
			regA |= sidMapper_readByte(regPC++);
			UPDATE_NZ(regA);
			cpuCycles += 2;
			break;

		case 0x05:		// ORA zp
			regA |= sidMapper_readByte(ZP_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 3;
			break;

		case 0x15:		// ORA zp,X
			regA |= sidMapper_readByte(ZPX_ADDR());
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x0D:		// ORA abs
			regA |= sidMapper_readByte(ABS_ADDR());
			regPC += 2;
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x1D:		// ORA abs,X
			ABSX_ADDR(addr);
			regA |= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x19:		// ORA abs,Y
			ABSY_ADDR(addr);
			regA |= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x01:		// ORA (zp,X)
			INDX_ADDR(addr);
			regA |= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 6;
			break;

		case 0x11:		// ORA (zp),Y
			INDY_ADDR(addr);
			regA |= sidMapper_readByte(addr);
			UPDATE_NZ(regA);
			cpuCycles += 5;
			break;

// == PHx ==
		case 0x48:		// PHA
			PUSHB(regA);
			cpuCycles += 3;
			break;

		case 0x08:		// PHP
			PUSHB(regF);
			cpuCycles += 3;
			break;

// == PLx ==
		case 0x68:		// PLA
			PULLB(regA);
			UPDATE_NZ(regA);
			cpuCycles += 4;
			break;

		case 0x28:		// PLP
			PULLB(regF);
			cpuCycles += 4;
			break;


// == ROL ==
		case 0x2A:		// ROL A
			ROL(regA);
			cpuCycles += 2;
			break;

		case 0x26:		// ROL zp
			addr = ZP_ADDR();
			operand = sidMapper_readByte(addr);
			ROL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 5;
			break;

		case 0x36:		// ROL zp,X
			addr = ZPX_ADDR();
			operand = sidMapper_readByte(addr);
			ROL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x2E:		// ROL abs
			addr = ABS_ADDR();
			regPC += 2;
			operand = sidMapper_readByte(addr);
			ROL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x3E:		// ROL abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ROL(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 7;
			break;

// == ROR ==
		case 0x6A:		// ROR A
			ROR(regA);
			cpuCycles += 2;
			break;

		case 0x66:		// ROR zp
			addr = ZP_ADDR();
			operand = sidMapper_readByte(addr);
			ROR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 5;
			break;

		case 0x76:		// ROR zp,X
			addr = ZPX_ADDR();
			operand = sidMapper_readByte(addr);
			ROR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x6E:		// ROR abs
			addr = ABS_ADDR();
			regPC += 2;
			operand = sidMapper_readByte(addr);
			ROR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 6;
			break;

		case 0x7E:		// ROR abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr);
			ROR(operand);
			sidMapper_writeByte(addr, operand);
			cpuCycles += 7;
			break;


// ====
		case 0x40:		// RTI
			PULLB(regF);
			PULLB(regPC);
			PULLB(addr);
			regPC += (addr << 8);
			cpuCycles += 6;
			break;

		case 0x60:		// RTS
			PULLB(regPC);
			PULLB(addr);
			regPC += (addr << 8) + 1;
			cpuCycles += 6;
			break;


// == SBC ==
		case 0xE9:		// SBC imm
			operand = sidMapper_readByte(regPC++) ^ 0xFF;
			ADC(operand);
			cpuCycles += 2;
			break;

		case 0xE5:		// SBC zp
			operand = sidMapper_readByte(ZP_ADDR()) ^ 0xFF;
			ADC(operand);
			cpuCycles += 3;
			break;

		case 0xF5:		// SBC zp,X
			operand = sidMapper_readByte(ZPX_ADDR()) ^ 0xFF;
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0xED:		// SBC abs
			operand = sidMapper_readByte(ABS_ADDR()) ^ 0xFF;
			regPC += 2;
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0xFD:		// SBC abs,X
			ABSX_ADDR(addr);
			operand = sidMapper_readByte(addr) ^ 0xFF;
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0xF9:		// SBC abs,Y
			ABSY_ADDR(addr);
			operand = sidMapper_readByte(addr) ^ 0xFF;
			ADC(operand);
			cpuCycles += 4;
			break;

		case 0xE1:		// SBC (zp,X)
			INDX_ADDR(addr);
			operand = sidMapper_readByte(addr) ^ 0xFF;
			ADC(operand);
			cpuCycles += 6;
			break;

		case 0xF1:		// SBC (zp),Y
			INDY_ADDR(addr);
			operand = sidMapper_readByte(addr) ^ 0xFF;
			ADC(operand);
			cpuCycles += 5;
			break;


// == SEx ==
		case 0x38:		// SEC
			regF |= EMU6502_FLAG_C;
			cpuCycles += 2;
			break;

		case 0xF8:		// SED
			regF |= EMU6502_FLAG_D;
			cpuCycles += 2;
			break;

		case 0x78:		// SEI
			regF |= EMU6502_FLAG_I;
			cpuCycles += 2;
			break;

// == STA ==
		case 0x85:		// STA zp
			addr = ZP_ADDR();
			sidMapper_writeByte(addr, regA);
			cpuCycles += 3;
			break;

		case 0x95:		// STA zp,X
			addr = ZPX_ADDR();
			sidMapper_writeByte(addr, regA);
			cpuCycles += 4;
			break;

		case 0x8D:		// STA abs
			addr = ABS_ADDR();
			regPC += 2;
			sidMapper_writeByte(addr, regA);
			cpuCycles += 4;
			break;

		case 0x9D:		// STA abs,X
			ABSX_ADDR(addr);
			sidMapper_writeByte(addr, regA);
			cpuCycles += 4;
			break;

		case 0x99:		// STA abs,Y
			ABSY_ADDR(addr);
			sidMapper_writeByte(addr, regA);
			cpuCycles += 4;
			break;

		case 0x81:		// STA (zp,X)
			INDX_ADDR(addr);
			sidMapper_writeByte(addr, regA);
			cpuCycles += 6;
			break;

		case 0x91:		// STA (zp),Y
			INDY_ADDR(addr);
			sidMapper_writeByte(addr, regA);
			cpuCycles += 5;
			break;


// == STX ==
		case 0x86:		// STX zp
			addr = ZP_ADDR();
			sidMapper_writeByte(addr, regX);
			cpuCycles += 3;
			break;

		case 0x96:		// STX zp,Y
			addr = ZPY_ADDR();
			sidMapper_writeByte(addr, regX);
			cpuCycles += 4;
			break;

		case 0x8E:		// STX abs
			addr = ABS_ADDR();
			regPC += 2;
			sidMapper_writeByte(addr, regX);
			cpuCycles += 4;
			break;

// == STY ==
		case 0x84:		// STY zp
			addr = ZP_ADDR();
			sidMapper_writeByte(addr, regY);
			cpuCycles += 3;
			break;

		case 0x94:		// STY zp,X
			addr = ZPX_ADDR();
			sidMapper_writeByte(addr, regY);
			cpuCycles += 4;
			break;

		case 0x8C:		// STY abs
			addr = ABS_ADDR();
			regPC += 2;
			sidMapper_writeByte(addr, regY);
			cpuCycles += 4;
			break;


// == Txx ==
		case 0xAA:		// TAX
			regX = regA;
			UPDATE_NZ(regX);
			cpuCycles += 2;
			break;

		case 0xA8:		// TAY
			regY = regA;
			UPDATE_NZ(regY);
			cpuCycles += 2;
			break;

		case 0xBA:		// TSX
			regX = regS;
			UPDATE_NZ(regX);
			cpuCycles += 2;
			break;

		case 0x8A:		// TXA
			regA = regX;
			UPDATE_NZ(regA);
			cpuCycles += 2;
			break;

		case 0x9A:		// TXS
			regS = regX;
			cpuCycles += 2;
			break;

		case 0x98:		// TYA
			regA = regY;
			UPDATE_NZ(regA);
			cpuCycles += 2;
			break;

// ====
		case 0xEA:		// NOP
		case 0x1A: case 0x3A: case 0x5A: case 0x7A: case 0xDA:
			cpuCycles += 2;
			break;

		default:
			ILLEGAL_OP();
			break;


		}
	}
}

